#!/usr/libexec/platform-python

import subprocess
import socket
import sys
import re
import os


class Colors:
    def __init__(self):
        # Regular colors
        self.red = '\033[91m'
        self.green = '\033[92m'
        self.yellow = '\033[93m'
        self.white = '\033[97m'
        self.reset = '\033[0m'

    def colorize(self, text, color):
        """Add color to text and handle reset automatically"""
        color_code = getattr(self, color.lower(), '')
        return f"{color_code}{text}{self.reset}"

def ok():
    colors = Colors()
    print(colors.colorize("[PASS]", "green"))

def fail(message):
    colors = Colors()
    print(colors.colorize("[FAIL]\n", "red"))
    print(colors.colorize("    " + message, "white"))

def warn(message):
    colors = Colors()
    print(colors.colorize("[WARN]\n", "yellow"))
    print(colors.colorize("    " + message, "white"))


def run_command(command):
    """Run a shell command and return the output"""
    try:
        result = subprocess.run(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True, universal_newlines=True)
        return result.stdout.strip()
    except subprocess.CalledProcessError as e:
        # Special handling for the dnf update --assumeno command
        if "dnf update --assumeno" or "is-enabled" or "is-active" in command:
            return e.stdout.strip()  # Return stdout (even if there was an error)
        else:
            # For other commands, keep the existing behavior
            return ""


def check_olvm_pre_check_script():
    """ Check if the oracle-ovirt-release-45 package is installed """
    print("+++ Checking oracle-ovirt-release-45 ", end="                    ", flush=True)
    script_package = "oracle-ovirt-release-45"
    check_package = run_command(f"rpm -qa | grep oracle-ovirt-release-45")

    if script_package in check_package:
        ok()
    else:
        fail("The 'oracle-ovirt-release-45-el8' package is not installed.\n    Please run: 'dnf install -y oracle-ovirt-release-45*'.\n")


def check_ovirt_engine_host_packages():
    """Check for the presence of the 'ovirt-engine' package"""
    print("+++ Checking if Host is installed ", end="                       ", flush=True)
    engine_package = "ovirt-engine"
    host_package = "ovirt-host"
    output = run_command(f"dnf list installed {engine_package} {host_package}")

    if engine_package in output and host_package in output:
        fail("FATAL: Both 'ovirt-engine' and 'ovirt-host' packages are installed.\n\n    - Never install the 'ovirt-engine' package in a KVM Host.\n    - Never add the Engine host in OLVM Admin Portal UI as a KVM Host.\n\n    Exiting the script: UNSUPPORTED CONFIGURATION !!!\n")
        sys.exit()
    elif engine_package in output:
        warn("The 'ovirt-engine' package is already installed.\n    DO NOT configure this Server as a KVM Host.\n")
        return True
    elif host_package in output:
        warn("The 'ovirt-host' package is already installed.\n    This is a KVM Host, DO NOT install the engine on this host.\n")
        return True
    else:
        ok(); return False


def check_minimal_install_group(installed):
    """Check if the 'Minimal Install' group is installed"""
    print("+++ Checking if a Minimal Installation ", end="                  ", flush=True)
    group_list_output = run_command("dnf grouplist --installed")
    
    if "Minimal Install" in group_list_output:
        ok(); return True
    else:
        warn("The 'Minimal Install' group was not found.\n    Please disregard this warning if:")
        print("      - This is a Hosted-Engine.\n      - This is a VM installed from an Oracle Template.\n")
        print("+++ As best effort, check package count ", end="                 ", flush=True)
        package_count = run_command("rpm -qa | wc -l")
        if int(package_count) < 500 or installed:
            ok(); return True
        else:
            fail("Too many packages already installed.\n    This is not a Minimal Installation as required.\n")
            return False


def get_minimal_install_info():
    """Fetch details of the 'Minimal Install' group"""
    minimal_install_info = run_command("dnf groupinfo 'Minimal Install'")
    mandatory_groups = []
    is_mandatory_section = False

    for line in minimal_install_info.splitlines():
        line = line.strip()
        if line.startswith("Mandatory Groups:"):
            is_mandatory_section = True
        elif line.startswith("Optional Groups:") or not line:
            is_mandatory_section = False
        elif is_mandatory_section:
            mandatory_groups.append(line)

    return mandatory_groups


def get_group_packages(group_name):
    """Retrieve mandatory and default packages for a group"""
    group_info = run_command(f"dnf groupinfo '{group_name}'")
    mandatory_packages, default_packages = [], []
    section = None

    for line in group_info.splitlines():
        line = line.strip()
        if line.startswith("Mandatory Packages:"):
            section = "mandatory"
        elif line.startswith("Default Packages:"):
            section = "default"
        elif line.startswith("Optional Packages:") or not line:
            section = None
        elif section == "mandatory":
            mandatory_packages.append(line.split("(")[0].strip())
        elif section == "default":
            default_packages.append(line.split("(")[0].strip())

    return mandatory_packages, default_packages


def check_packages_installed(packages):
    """Verify if specific packages are installed"""
    missing_packages = []
    exclude_packages = "dnf-plugin-spacewalk, plymouth, rhn-client-tools, rhn-setup, rhnlib, rhnsd"
    for package in packages:
        if package not in exclude_packages:
            rpm_check = run_command(f"rpm -q {package}")
            if "is not installed" in rpm_check or not rpm_check.strip():
                missing_packages.append(package)
    return missing_packages


def validate_minimal_install():
    """Validate the 'Minimal Install' group by checking mandatory packages"""
    print("+++ Validating the 'Minimal Install' Group ", end="              ", flush=True)
    mandatory_groups = get_minimal_install_info()

    if not mandatory_groups:
        warn("No mandatory groups found.\n")
        return
    
    missing_packages = []
    for group in mandatory_groups:
        mandatory_packages, _ = get_group_packages(group)
        missing_packages.extend(check_packages_installed(mandatory_packages))

    if not missing_packages:
        ok()
    else:
        warn("If using ULN or local repos, the following packages are missing:")
        print(f"    {', '.join(missing_packages)}\n")


def check_enabled_repos():
    """Validate enabled repositories"""
    print("+++ Checking enabled repositories ", end="                       ", flush=True)
    systemid = '/etc/sysconfig/rhn/systemid'
    uln = os.path.exists(systemid)
    repo_list_output = run_command("dnf repolist enabled")
    if uln:
        enabled_repos = [line.split()[0] for line in repo_list_output.splitlines()[2:]]
        required_repos = [
            "ol8_x86_64_baseos_latest", "ol8_x86_64_appstream", "ol8_x86_64_kvm_appstream", "ol8_x86_64_addons", 
            "ol8_x86_64_ovirt45", "ol8_x86_64_ovirt45_extras", "ol8_x86_64_gluster_appstream", "ol8_x86_64_UEKR7"
        ]
    else:
        enabled_repos = [line.split()[0] for line in repo_list_output.splitlines()[1:]]
        required_repos = [
            "ol8_baseos_latest", "ol8_appstream", "ol8_kvm_appstream", "ol8_addons", 
            "ovirt-4.5", "ovirt-4.5-extra", "ol8_gluster_appstream", "ol8_UEKR7"
        ]
    
    extra_repos = [repo for repo in enabled_repos if repo not in required_repos]
    missing_repos = [repo for repo in required_repos if repo not in enabled_repos]

    if not extra_repos and not missing_repos:
        ok()
    else:
        if extra_repos:
            warn("Extra repositories are enabled:")        
            print(f"    {', '.join(extra_repos)}\n")
            if not uln:
                print("    Please run the command:")
                print(f"    dnf config-manager --set-disabled {' '.join(extra_repos)}\n")
            else:
                print("    Please disable the following ULN Channels:")
                print(f"    {' '.join(extra_repos)}\n")
        if missing_repos and extra_repos:
            print("+++ Checking missing repositories ", end="                       ", flush=True)
            fail("Missing repositories:")           
            print(f"    {', '.join(missing_repos)}\n")
            if not uln:
                print("    Please run the command:")
                print(f"    dnf config-manager --set-enabled {' '.join(missing_repos)}\n")
            else:
                print("    Please enable the following ULN Channels:")
                print(f"    {' '.join(missing_repos)}\n")
        elif missing_repos:
            fail("Missing repositories:")           
            print(f"    {', '.join(missing_repos)}\n")
            if not uln:
                print("    Please run the command:")
                print(f"    dnf config-manager --set-enabled {' '.join(missing_repos)}\n")
            else:
                print("    Please enable the following ULN Channels:")
                print(f"    {' '.join(missing_repos)}\n")           


def check_repositories():
    """Validate repository reachability using dnf makecache and dry-run update."""
    print("+++ Running 'dnf makecache' ", end="                             ", flush=True)
    # Check with `dnf makecache`
    makecache_output = run_command("dnf makecache")

    if "Metadata cache created" in makecache_output:
        ok()
    else:
        warn("Issues encountered when running 'dnf makecache':\n")        
        print(    makecache_output)

    # Dry-run update with `dnf update --assumeno`
    print("+++ Dry run 'dnf update --assumeno' ", end="                     ", flush=True)
    dry_run_output = run_command("sudo dnf update --assumeno")

    if "Dependencies resolved." in dry_run_output:
        ok()
    else:
        warn("Errors or warnings encountered during dry-run update:\n")    
        print(    dry_run_output)


def parse_version(version_string):
    """Parse version string into a tuple of integers (major, minor, patch)."""
    match = re.match(r"(\d+)\.(\d+)\.(\d+)", version_string)
    if match:
        return tuple(map(int, match.groups()))
    return None


def check_ansible_version():
    """Verify Ansible installation and version"""
    print("+++ If installed, check ansible version ", end="                 ", flush=True)
    output = run_command("ansible --version")
    if not output:
        ok()
        return
    
    match = re.search(r"ansible \[core (\d+\.\d+\.\d+)\]", output)
    if match:
        version = match.group(1)
        min_version, max_version = (2, 16, 0), (2, 16, 99)
        parsed_version = tuple(map(int, version.split(".")))
        if min_version <= parsed_version <= max_version:
            ok()
        else:
            fail("Ansible version is outside the supported range.\n    When installed, ansible-core must be >= 2.16.\n")   
    else:
        warn("Unable to parse Ansible version.\n")


def check_firewalld_status():
    """Check if the firewalld service is running."""
    print("+++ Checking Firewalld status ", end="                           ", flush=True)
    fstatus = run_command("systemctl is-active firewalld")
    fenable = run_command("systemctl is-enabled firewalld")

    if fstatus == "active" and fenable == "enabled":
        ok();   
    elif fstatus == "active" and fenable == "disabled":
        warn("Firewalld is running, but not enabled.\n    Please run 'systemctl enable firewalld'.\n")
    elif fstatus == "inactive" and fenable == "enabled":
        warn("Firewalld is enabled, but not running.\n    Please run 'systemctl start firewalld'\n")
    elif fstatus == "inactive" and fenable == "disabled":
        warn("Firewalld is disabled, and it is not running.\n    Please run 'systemctl enable firewalld; systemctl start firewalld'\n")
    else:
        fail("Unable to check firewalld status.\n")


def check_selinux_status():
    """Check and report the current SELinux status."""
    print("+++ Checking SELinux status ", end="                             ", flush=True)

    try:
        sestatus_output = run_command("sestatus")
        if not sestatus_output:
            fail("SELinux might not be installed or configured.\n")
            return

        if "enabled" in sestatus_output:
            if "enforcing" in sestatus_output:
                ok()
            else:
                warn("SELinux is enabled but not in enforcing mode.\n")
        else:
            warn("SELinux is not enabled.\n")
    except Exception as e:
        fail("Error checking SELinux status.\n")


def check_fips_status():
    """Check if FIPS mode is enabled on the system."""
    print("+++ Checking FIPS status ", end="                                ", flush=True)
    fips_enabled = False

    # Check FIPS via sysctl
    try:
        fips_status = run_command("sysctl -n crypto.fips_enabled")
        if fips_status == "1":
            fips_enabled = True
    except Exception:
        fail("Error checking FIPS status via sysctl.\n")
        return

    # Check FIPS via /proc/sys/crypto/fips_enabled
    try:
        with open("/proc/sys/crypto/fips_enabled", "r") as f:
            if f.read().strip() == "1":
                fips_enabled = True
    except FileNotFoundError:
        fail("FIPS check might not apply to your system.\n")
        return

    # Output FIPS status
    if fips_enabled:
        ok(); print("    FIPS is enabled.")
    else:
        ok(); print("    FIPS is disabled.")


def check_kernel_version():
    """ Check kernel version, recommend UEK """
    print("+++ Checking Linux Kernel ", end="                               ", flush=True)
    uek7_kernel = "5.15.0-"
    inst_kernel = run_command(f"uname -r")
    if uek7_kernel in inst_kernel:
        ok()
        """ Check if uek-modules-extra is installed """
        print("+++ Checking kernel-uek-modules-extra ", end="                   ", flush=True)
        uek7_extra = "kernel-uek-modules-extra"
        inst_extra = run_command(f"rpm -qa | grep kernel-uek-modules-extra")
        if uek7_extra in inst_extra:
            ok()
        else:
            warn("Package kernel-uek-modules-extra is not installed.\n    Please run: 'dnf install -y kernel-uek-modules-extra'\n")
    else:
        warn("The host is not running Oracle UEK7 (Recommended).\n")


def check_qemu_version():
    """Verify qemu version"""
    print("+++ If installed, check qemu-kvm version ", end="                ", flush=True)
    output = run_command("dnf list installed qemu-kvm")
    if 'qemu-kvm' not in output:
        ok()
    elif '7.2.0' in output:
        ok()
    else:
        fail("The qemu-kvm package version installed in the system is not supported.\n    The correct qemu-kvm version for this release is >= 7.2.0\n    Please re-install the host following OLVM Getting Started Guide instructions:")
        print ("    https://docs.oracle.com/en/virtualization/oracle-linux-virtualization-manager/getstart/index.html\n")

   
def check_libvirt_version():
    """Verify libvirt version"""
    print("+++ If installed, check libvirt version ", end="                 ", flush=True)
    output = run_command("dnf list installed libvirt")
    if 'libvirt' not in output:        
        ok()
    elif '9.0.0' in output:
        ok()
    else:
        fail("The libvirt package version installed in the system is not supported.\n    The correct libvirt version for this release is >= 9.0.0\n    Please re-install the host following OLVM Getting Started Guide instructions:")
        print ("    https://docs.oracle.com/en/virtualization/oracle-linux-virtualization-manager/getstart/index.html\n")


def check_hostname():
    """ verify if FQDN exists """
    print("+++ Checking Hostname/FQDN ", end="                              ", flush=True)
    hname = socket.getfqdn()
    fqdn = socket.gethostname()

    if "." not in hname:
        fail("The system does not have a valid FQDN.\n    It is recommended to use an external DNS Server for Name Resolution.\n    As a last resort, the file '/etc/hosts' can also be used.")
    else:
        if hname == fqdn:
            ok()
        else:
            output = run_command(f"grep {hname} /etc/hosts")
            if output:
                warn("The FQDN is coming from the system /etc/hosts file.\n    It is recommended to use an external DNS Server for Name Resolution.")
            else:
                """ The FQDN exists, but it is not possible to determine where it is coming from """
                ok()

def main():
    print("\n-----------------------------------\n    OLVM 4.5.5 PRE-CHECK SCRIPT\n-----------------------------------\n")
    check_olvm_pre_check_script()
    installed = check_ovirt_engine_host_packages()
    check_minimal_install_group(installed)
    validate_minimal_install()
    check_enabled_repos()
    check_repositories()
    check_kernel_version()
    check_firewalld_status()
    check_selinux_status()
    check_fips_status()
    check_ansible_version()
    check_qemu_version()
    check_libvirt_version()
    check_hostname()
    print()

if __name__ == "__main__":
    main()


